# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Arm Ltd. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
#

import inspect
import logging
import re
import sys

from argparse import ArgumentParser, SUPPRESS, Namespace
from enum import Enum
from gettext import gettext as _
from inspect import signature, Signature
from itertools import islice
from types import FunctionType

from typing import _GenericAlias as GenericAlias

from arm.avhclient import AvhClient, AvhBackend, __version__


class AvhCli:
    """Arm Virtual Hardware Command Line Interface"""

    def __init__(self):
        parser = self._parser()

        args = parser.parse_known_args()[0]
        if args.verbosity:
            verbosity = args.verbosity
            logging.basicConfig(format='[%(levelname)s]\t%(message)s', level=verbosity)
            logging.debug("Verbosity level is set to %s", verbosity)

        avh_client = AvhClient(args.backend)

        self._add_commands(parser)
        self._add_backend_args(parser, avh_client.backend)

        args = parser.parse_args()

        self._consume_backend_args(avh_client.backend, args)

        func = AvhClient.__dict__[args.subcmd.replace('-', '_')]
        params = signature(func).parameters
        func_args = [vars(args)[param.replace('-', '_')] for param in islice(params.keys(), 1, None)]
        try:
            func(avh_client, *func_args)
        except RuntimeError as e:
            if e.__cause__ is not None:
                logging.error(e.__cause__.__doc__)
                logging.error(e.__cause__)
            if str(e):
                logging.error(e)
            sys.exit(1)
        sys.exit(0)

    @staticmethod
    def _parser() -> ArgumentParser:
        parser = ArgumentParser(add_help=False)

        parser.add_argument('-v', '--verbosity',
                            type=str,
                            choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
                            default='INFO',
                            help='Set the output verbosity. Default: INFO')

        parser.add_argument('-b', '--backend',
                            type=str,
                            choices=AvhClient.get_available_backends(),
                            default='aws',
                            help=f'Select AVH backend to use. Default: {AvhClient.get_available_backends()[0]}')

        parser.add_argument('--version', action='version', version=f'%(prog)s {__version__}')

        return parser

    @staticmethod
    def _add_argument(parser, argname: str, argtype: type = str, default=None, helptext: str = ''):
        kwargs = {'help': helptext, 'required': default is None}
        if argtype is bool:
            kwargs['action'] = 'store_true'
        elif isinstance(argtype, GenericAlias):
            if argtype.__origin__ is list:
                kwargs['type'] = argtype.__args__[0] if len(argtype.__args__) > 0 else str
                kwargs['nargs'] = '+' if default is None else '*'
        elif issubclass(argtype, Enum):
            kwargs['choices'] = list(filter(lambda v: v.value, argtype))
            kwargs['type'] = type(kwargs['choices'][0])
        else:
            kwargs['type'] = argtype
        if default is not None:
            kwargs['default'] = default
            if not kwargs['help'].endswith('\n'):
                kwargs['help'] += '\n'
            kwargs['help'] += f"Defaults to '{default}'."
        parser.add_argument(f"--{argname.replace('_', '-')}", **kwargs)

    @staticmethod
    def _add_backend_args(parser: ArgumentParser, backend: AvhBackend):
        group = parser.add_argument_group(f"{backend.name()} backend properties")
        for key, value in filter(lambda m: not m[0].startswith('_') and isinstance(m[1], property),
                           backend.__class__.__dict__.items()):
            sig = inspect.signature(value.fget)
            AvhCli._add_argument(group, key, sig.return_annotation, getattr(backend, key), value.__doc__)

    @staticmethod
    def _consume_backend_args(backend: AvhBackend, args: Namespace):
        args = vars(args)
        for key, _ in filter(lambda m: not m[0].startswith('_') and isinstance(m[1], property),
                           backend.__class__.__dict__.items()):
            if key in args:
                setattr(backend, key, args[key])

    @staticmethod
    def _add_commands(parser: ArgumentParser):
        parser.add_argument('-h', '--help', action='help',
                            default=SUPPRESS, help=_('show this help message and exit'))

        subparsers = parser.add_subparsers(dest='subcmd', required=True, help='sub-command help')

        for name, member in AvhClient.__dict__.items():
            if isinstance(member, FunctionType) and not name.startswith('_'):
                func_help = member.__doc__.split('\n')[0] if member.__doc__ else ''
                subparser = subparsers.add_parser(name.replace('_', '-'), help=func_help)
                params = signature(member).parameters
                for param in islice(params.items(), 1, None):
                    param_help = re.search(f"{param[0]}: (.*)", member.__doc__) if member.__doc__ else None
                    param_help = param_help.group(1) if param_help else ""
                    param_type = param[1].annotation if param[1].annotation != Signature.empty else str
                    param_default = param[1].default if param[1].default != Signature.empty else None
                    AvhCli._add_argument(subparser, param[0], param_type, param_default, param_help)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Arm Ltd. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
#

import logging
import os
import tarfile

from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List

import yaml

from .avh_backend import AvhBackend, AvhBackendState
from .helper import create_archive


class AvhSpec:
    """Wrapper for AVH spec file."""

    def __init__(self, path: Path):
        """AVH spec file

        Format of the specfile:
            name: (optional) The name of the workload.
            workdir: (optional) The local directory to use as the workspace, defaults to specfile's parent.
            backend: (optional) Dictionary with backend specific parameters.
              aws: (optional) Dictionary with AWS backend specific parameters. (see backend help)
              local: (optional) Dictionary with local backend specific parameters. (see backend help)
            upload: (optional) List of glob patterns of files to be sent to the AVH backend. (see glob format)
            steps: (mandatory) List of steps to be executed on the AVH backend.
              - run: String written into a bash script and executed on the AVH backend inside the workspace directory.
            download: (optional) List of glob patterns of files to be retrieved back from the AVH backend.
               (see glob format)

        Glob format:
            The list of glob patterns is evaluated in order.
            Wildcard '*' matches all files but no directory except hidden files (starting with '.').
            Wildcard '**' matches all files and directories except hidden files/directories (starting with '.').
            Inclusive matches (no prefix) are added to the file list.
            Exclusive (prefixed with '-:') matches are removed from current file list.
        """
        self._path = path

        if not path.exists():
            raise RuntimeError from FileNotFoundError(path)

        with open(path, encoding='UTF-8') as file:
            self._spec = yaml.safe_load(file)

    def backend_settings(self, backend: str) -> Dict[str, Any]:
        """Get the backend specific settings.

        Params:
            backend: The backend name to get settings for.

        Returns:
            The mapping of settings for the given backend.
        """
        return self._spec.get('backend', {}).get(backend, {})

    @property
    def workdir(self) -> Path:
        """The working directory within the local workspace."""
        return self._path.parent.joinpath(self._spec.get('workdir', '.')).resolve()

    @property
    def upload(self) -> List[str]:
        """The glob pattern for files to be uploaded to the backend."""
        return self._spec.get('upload', ['**/*'])

    @property
    def steps(self) -> List:
        """The steps to be executed on the backend."""
        return self._spec.get('steps', [])

    @property
    def download(self) -> List[str]:
        """The glob pattern for files to be downloaded from the backend."""
        return self._spec.get('download', ['**/*'])


class AvhClient:
    """AVH Client"""

    @staticmethod
    def get_available_backends() -> List[str]:
        """Get a list of available backend implementations.

        Returns:
            List of backend names sorted by priority.
        """
        backends = AvhBackend.find_implementations()
        return sorted(backends.keys(), key=lambda k: backends[k].priority())

    def __init__(self, backend):
        self.backend_desc = backend.lower()
        logging.info(f"avh:{self.backend_desc} backend selected!")
        self._set_backend()

    def _set_backend(self):
        backends = AvhBackend.find_implementations()
        if self.backend_desc in backends:
            self.backend = backends[self.backend_desc]()
        else:
            logging.error(f"{self.backend_desc} not supported!")
            raise RuntimeError()

    def prepare(self, force: bool = False) -> AvhBackendState:
        """Prepare the backend to execute AVH workload.

        Args:
            force: Force (re)preparation of the backend.

        Returns:
            The BackendState the backend was in before preparation.
        """
        return self.backend.prepare()

    def upload(self, workspace: Path = Path(""), patterns: List[str] = None):
        """Upload the workspace to the AVH backend

        Args:
            workspace: The base directory for the workspace
            patterns: List if glob patters. Patterns prefixed with -: denote excludes.
        """
        if not patterns:
            patterns = ['**/*']
        avhin = None
        try:
            avhin = NamedTemporaryFile(mode='w+b', prefix='avhin-', suffix='.tbz2', delete=False)
            avhin.close()
            create_archive(avhin.name, workspace, patterns, verbose=True)
            self.backend.upload_workspace(avhin.name)
        finally:
            if avhin:
                os.remove(avhin.name)

    def run(self, cmds: List[str]):
        """Run given commands on AVH backend

        Args:
            cmds: List of commands to run.
        """
        self.backend.run_commands(cmds)

    def download(self, workspace: Path = Path(""), patterns: List[str] = None):
        """Download the workspace from the AVH backend

        Args:
            workspace: The base directory for the workspace
            patterns: List if glob patters. Patterns prefixed with -: denote excludes.
        """
        if not patterns:
            patterns = ['**/*']
        avhout = None
        try:
            avhout = NamedTemporaryFile(mode='r+b', prefix='avhout-', suffix='.tbz2', delete=False)
            avhout.close()
            self.backend.download_workspace(avhout.name, patterns)
            with tarfile.open(avhout.name, mode='r:bz2') as archive:
                archive.list(verbose=False)
                archive.extractall(path=workspace)
        finally:
            if avhout:
                os.remove(avhout.name)

    def cleanup(self, state: AvhBackendState = AvhBackendState.CREATED):
        """Cleanup backend into a former state.

        Args:
            state: The state to turn backend into.
        """
        self.backend.cleanup(state)

    def execute(self, specfile: Path = Path("./avh.yml")):
        """Execute the AVH job specified by given specfile.

        Args:
            specfile: Path to the YAML specfile.
        """
        backend_state = AvhBackendState.INVALID

        spec = AvhSpec(specfile)
        for key, value in spec.backend_settings(self.backend.name()).items():
            kkey = key.replace('-', '_')
            if hasattr(self.backend, kkey):
                setattr(self.backend, kkey, value)

        try:
            logging.info("")
            logging.info("Preparing instance...")
            logging.info('='*80)
            backend_state = self.backend.prepare()

            logging.info("")
            logging.info("Uploading workspace...")
            logging.info('='*80)
            self.upload(spec.workdir, spec.upload)

            logging.info("")
            logging.info("Executing...")
            logging.info('='*80)
            for step in spec.steps:
                if 'run' in step:
                    cmds = [cmd for cmd in step['run'].split('\n') if cmd]
                    self.run(cmds)

            logging.info("")
            logging.info("Downloading workspace...")
            logging.info('='*80)
            self.download(spec.workdir, spec.download)
        finally:
            logging.info("")
            logging.info("Teardown instance...")
            logging.info('='*80)
            self.backend.cleanup(backend_state)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Arm Ltd. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
#

import tarfile
from glob import iglob

from pathlib import Path
from typing import Union, List, Iterable


def _iglob(pathname: Union[str, Path], root_dir: Union[str, Path] = Path.cwd(),
           recursive: bool = True, files_only: bool = True) -> Iterable[Path]:
    """Wrapper for glob.iglob

    Args:
        pathname: A glob pattern to evaluate relative to root_dir.
        root_dir: The base directory to evaluate pattern from, defaults to cwd.
        recursive: Make ** matching subdirectories, default to true.
        files_only: Returns only normal files

    Returns:
        Generator listing all match results.
    """
    if not isinstance(root_dir, Path):
        root_dir = Path(root_dir)
    pathname = root_dir.joinpath(pathname)
    for match in iglob(str(pathname), recursive=recursive):
        match = Path(match)
        if not files_only or match.is_file():
            yield match


def create_archive(filename: Union[str, Path],
                   root_dir: Union[str, Path] = Path.cwd(),
                   globs: List[str] = None,
                   verbose: bool = False):
    """Create an bzip2-compressed tarball of the given directory.

    Files matching the given glob patterns underneath root_dir are archived. The patterns
    are applied in order. Patterns without prefix denote includes. A pattern prefixed with
    `-:` denotes an exclude and removes all matching files that have been included previously.

    Args:
        filename: The filename of the resulting archive.
        root_dir: The root directory for the archive, defaults to current working directory.
        globs: A list of glob filters with includes and excludes, defaults to **/*.
        verbose: List archive content if set to True.
    """
    if not isinstance(root_dir, Path):
        root_dir = Path(root_dir)
    if not globs:
        globs = ["**/*"]

    with tarfile.open(filename, mode='w:bz2') as archive:
        files = set()
        for pattern in globs:
            if pattern.startswith('-:'):
                files = files - set(_iglob(pattern[2:], root_dir=root_dir))
            else:
                files.update(set(_iglob(pattern, root_dir=root_dir)))
        for file in files:
            archive.add(file, arcname=file.relative_to(root_dir))
        if verbose:
            archive.list(verbose=False)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Arm Ltd. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
#

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Dict, Type, List, Union


class AvhBackendState(Enum):
    """Possible AWS EC2 instance states."""
    INVALID = None
    CREATED = 'created'
    STARTED = 'started'
    RUNNING = 'running'

    def __str__(self):
        return self.value


class AvhBackend:
    """Backend interface"""

    @staticmethod
    def find_implementations() -> Dict[str, Type[AvhBackend]]:
        """Find all available backend implementations.

        Returns:
            Mapping with backend names and classes.
        """
        return {cls.name(): cls for cls in AvhBackend.__subclasses__()}

    @staticmethod
    def name() -> str:
        """Return the name this backend shall be published as.

        Returns:
            Unique backend identifier.
        """
        raise NotImplementedError()

    @staticmethod
    def priority() -> int:
        """Return a priority fir this backend.
        The priority defines the order different backend implementations are offered to the user.
        The lower the priority value the higher the backend will be listed.

        Returns:
            Priority for this backend
        """
        raise NotImplementedError()

    def prepare(self, force: bool = False) -> AvhBackendState:
        """Runs required commands to prepare the backend for AVH workload.

        Params:
            force: Force preparation on already prepared backends.

        Returns:
            The BackendState the backend was in before preparation.
         """
        raise NotImplementedError()

    def cleanup(self, state: AvhBackendState):
        """Cleanup the backend.
        The backend is brought back into state before call to prepare.

        Params:
            state: The state returned by prepare
        """
        raise NotImplementedError()

    def upload_workspace(self, filename: Union[str, Path]):
        """Upload the workspace content from the given tarball.

        Params:
            filename: The archived workspace.
        """
        raise NotImplementedError()

    def download_workspace(self, filename: Union[str, Path], globs: List[str] = None):
        """Download the workspace content into given tarball.

        Params:
            filename: The archived workspace.
        """
        raise NotImplementedError()

    def run_commands(self, cmds: List[str]):
        """Execute the given commands on the backend.

        Params:
            cmds - List of command strings
        """
        raise NotImplementedError()
